unit sprites;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  DGCILib, DGC;

{ Sprite routines by Daniel Parnell August 1996 }

const
  hasBeenKilled = $8000;  // Tells the engine to destroy a sprite
  wantBounce    = $4000;  // Makes a sprite bounce at the edge of it's bounds
  wantStick     = $2000;  // Makes the edges "sticky"

type
  TSpriteEngine = class;
  TSprite       = class;

  { The TDirtyRect object keeps track of those portions of the screen that
    need to be updated (ie those that are "dirty") }
  TDirtyRect = class(TObject)
  private
    next  : TDirtyRect;
    count : integer;

    function GetRect:TRect;
    procedure SetRect(value : TRect);
  public
    a,b  : TPoint;

    property Rect : TRect read GetRect write SetRect;

    function Intersects(r : TDirtyRect):boolean;
    procedure Assign(ax,ay,bx,by : integer);
    function Contains(p : TPoint):boolean;
    procedure Union(r : TDirtyRect);
  end;

  TEdge = (TopEdge,LeftEdge,BottomEdge,RightEdge);
  TEdges = set of TEdge;

  { The TSprite object represents a moveable image that can be placed
    anywhere on the screen and move about without disturbing the background
    or other sprites }
  TSprite=class(TObject)
  private
    owner       : TSpriteEngine;
    fx,fy       : integer;      { Position of this sprite }
    fvx,fvy,
    fax,fay     : integer;

    fImage      : integer;      { The image that this sprite displays }
    fVisible    : boolean;      { True if the sprite should be visible }
    fMask       : integer;      { Collision mask }

    procedure SetImage(value : integer);
    procedure SetInfo(index : integer; value : integer);
    function  GetInfo(index : integer):integer;
    procedure SetVisible(value : boolean);
  public
    next,prev : TSprite;
    bounds    : TRect;     { Used to confine a sprite to an area }
    flags     : integer;   { Misc flags available to the user }

    property x       : integer index 1 read GetInfo  write SetInfo;
    property y       : integer index 2 read GetInfo  write SetInfo;
    property vx      : integer index 3 read GetInfo  write SetInfo;
    property vy      : integer index 4 read GetInfo  write SetInfo;
    property ax      : integer index 5 read GetInfo  write SetInfo;
    property ay      : integer index 6 read GetInfo  write SetInfo;
    property Image   : integer         read fImage   write SetImage;
    property Visible : boolean         read fVisible write SetVisible;
    property Mask    : integer         read fMask    write fMask;

    constructor Create(AOwner : TSpriteEngine); virtual;
    procedure   MoveTo(ax,ay : integer);
    procedure   MoveRel(dx,dy : integer);
    procedure   Show;
    procedure   Hide;
    procedure   Invalidate;
    procedure   Tick; virtual;
    procedure   Collide(s : TSprite); virtual;
    procedure   Kill; virtual;
    procedure   HitEdge(e : TEdges); virtual;
    destructor  Destroy; override;
  end;

  { The TBackground object is used to place a picture in the background of
    the screen.  This can be used to give the sprites something to move about
    on. It is basically just a sprite that fills the entire display. }
  TBackground=class(TSprite)
    constructor Create(AOwner : TSpriteEngine; aImage : integer);
  end;

  { A sprite that uses floating point numbers to store much of it's information.
    This is useful for making sprites that move at any angle }
  TFractSprite = class(TSprite)
    rx,ry,rvx,rvy : single;

    procedure Tick; override;
  end;

  TAnimSprite = class(TSprite)
  private
    count     : integer;
  public
    imageBase : integer;
    numFrames : integer;
    speed     : integer;

    constructor Create(AOwner : TSpriteEngine; base,frames : integer);
    procedure Tick; override;
  end;

  TSpriteCollision=procedure(Sender : TSpriteEngine; s1,s2  : TSprite) of object;

  { TSpriteEngine keeps track of the dirty rectangles on the screen and redraws
    only those portions of the screen that are necessary }

  TSpriteEngine = class(TComponent)
  private
    dirty        : TDirtyRect;      { List of the dirty rectangles }
    fSprites     : TSprite;         { A doubly linked list of sprites }
    fBackground  : TBackground;     { A pointer to the background sprite }
    fScreen      : TDGCScreen;
    fOnCollision : TSpriteCollision;

    procedure AddDirtyRect(x,y,w,h : word);
    procedure RemoveDirtyRect(r : TDirtyRect);
    procedure MergeDirtyList;
    procedure EmptyDirtyList;
    procedure DrawSprite(s : TSprite; r : TDirtyRect);
    function  BeforeSprite(Sprite : TSprite):TSprite;
  public
    property Sprites    : TSprite     read fSprites;
    property Background : TBackground read fBackground;

    constructor Create(AOwner : TComponent); override;
    destructor  Destroy; override;
    procedure   UpdateSprites;
  published
    property Screen      : TDGCScreen       read fScreen      write fScreen;
    property OnCollision : TSpriteCollision read fOnCollision write fOnCollision;
  end;

procedure Register;

implementation
uses trace;

function Min(a,b : integer):integer;
begin
  if a<b then
    result:=a
  else
    result:=b;
end;

function Max(a,b : integer):integer;
begin
  if a>b then
    result:=a
  else
    result:=b;
end;

function TDirtyRect.GetRect:TRect;
begin
  Result.TopLeft:=a;
  Result.BottomRight:=b;
end;

procedure TDirtyRect.SetRect(value : TRect);
begin
  a:=value.TopLeft;
  b:=value.BottomRight;
end;

procedure TDirtyRect.Union(r : TDirtyRect);
begin
  a.x:=min(a.x,r.a.x);
  a.y:=min(a.y,r.a.y);
  b.x:=max(b.x,r.b.x);
  b.y:=max(b.y,r.b.y);
  count:=2;
end;

function TDirtyRect.Contains(p : TPoint):boolean;
begin
  result:=(p.x>=a.x) and (p.x<=b.x) and
          (p.y>=a.y) and (p.y<=b.y);
end;

procedure TDirtyRect.Assign(ax,ay,bx,by:integer);
begin
  count:=2;

  a.x:=ax;
  a.y:=ay;
  b.x:=bx;
  b.y:=by;
end;

function TDirtyRect.Intersects(r : TDirtyRect):boolean;
var
  p,q,w,e : TPoint;

begin
  p.x:=b.x; p.y:=a.y;
  q.x:=a.x; q.y:=b.y;

  w.x:=r.b.x; w.y:=r.a.y;
  e.x:=r.a.x; e.y:=r.b.y;

  if contains(r.a) or contains(r.b) or
     contains(w)   or contains(e) or
     r.contains(a) or r.contains(b) or
     r.contains(p) or r.contains(q) then
    result:=TRUE
  else
    result:=FALSE;
end;

procedure TSpriteEngine.AddDirtyRect(x,y,w,h : word);
var
  r         : TDirtyRect;
  temp,last : TDirtyRect;

begin
  r:=TDirtyRect.Create;
  r.assign(x,y,x+w,y+h);
  r.count:=2;

  temp:=dirty;

  if temp<>nil then
    begin
      while temp<>nil do
        begin
          last:=temp;
          temp:=temp.next;
        end;

      temp:=r;
      last.next:=temp;
    end
  else
    begin
      temp:=r;
      dirty:=temp;
    end;

  temp.next:=nil;
end;

procedure TSpriteEngine.RemoveDirtyRect(r : TDirtyRect);
var
  temp,last : TDirtyRect;

begin
  { By default a rectangle stays dirty for two frames so that
    it will be drawn on both pages of the display }
  dec(r.count);

  if r.count>0 then
    exit;

  temp:=dirty;
  last:=nil;
  while (temp<>nil) and (temp<>r) do
    begin
      last:=temp;
      temp:=temp.next;
    end;

  if last=nil then
    dirty:=r.next
  else
    last.next:=r.next;

  r.Free;
end;

{ This routine goes through the list of dirty rectangles and combines any that
  overlap.  This reduces the number of operations that need to be performed}
procedure TSpriteEngine.MergeDirtyList;
var
  t1,t2 : TDirtyRect;
  flag  : boolean;

begin
  t1:=dirty;
  while t1<>nil do
    begin
      flag:=FALSE;
      t2:=dirty;

      while (t2<>nil) and not flag do
        begin
          if (t1<>t2) and t1.Intersects(t2) then
            begin
              t1.union(t2);
              t2.count:=0;
              RemoveDirtyRect(t2);
              flag:=TRUE;
            end
          else
            t2:=t2.next;
        end;

      if not flag then
        t1:=t1.next;
    end;
end;

procedure TSpriteEngine.EmptyDirtyList;
var
  temp,t2 : TDirtyRect;

begin
  temp:=dirty;
  while temp<>nil do
    begin
      t2:=temp.next;
      RemoveDirtyRect(temp);
      temp:=t2;
    end;
end;

{ Draws a sprite to the screen while clipping it to the rectabgle given by r }
procedure TSpriteEngine.DrawSprite(s : TSprite; r : TDirtyRect);
var
  rec : TRect;

  procedure ConstrainRect(var r1 : TRect; r2 : TRect);
  var
    temp : integer;

  begin
    if r1.top<r2.top then
      r1.top:=r2.top;
    if r1.top>r2.bottom then
      r1.top:=r2.bottom;
    if r1.left<r2.left then
      r1.left:=r2.left;
    if r1.left>r2.right then
      r1.left:=r2.right;
    if r1.bottom<r2.top then
      r1.bottom:=r2.top;
    if r1.bottom>r2.bottom then
      r1.bottom:=r2.bottom;
    if r1.right<r2.left then
      r1.right:=r2.left;
    if r1.right>r2.right then
      r1.right:=r2.right;

    if r1.top>r1.bottom then
      begin
        temp:=r1.top;
        r1.top:=r1.bottom;
        r1.bottom:=temp;
      end;

    if r1.left>r1.right then
      begin
        temp:=r1.left;
        r1.left:=r1.right;
        r1.right:=temp;
      end;
  end;

begin
  rec:=r.Rect;
  constrainRect(rec,rect(0,0,fScreen.ScreenWidth,fScreen.ScreenHeight));

  if s.fImage>=0 then
    begin
      rec.top:=rec.top-s.fy;
      rec.left:=rec.left-s.fx;
      rec.bottom:=rec.bottom-s.fy;
      rec.right:=rec.right-s.fx;

      ConstrainRect(rec,rect(0,0,fScreen.images[s.fImage].width,
                                 fScreen.images[s.fImage].height));
      if (rec.left<>rec.right) and (rec.top<>rec.bottom) then
       fScreen.Back.BltFast(s.fx+rec.left,s.fy+rec.top,fScreen.Images[s.fimage],rec,TRUE);
    end
  else
    fScreen.Back.EraseRect(rec,(abs(s.fImage)-1) and $ff);
end;

{ Returns the sprite that preceeds the given sprite in the sprite list }
function TSpriteEngine.BeforeSprite(Sprite : TSprite):TSprite;
var
  last,temp : TSprite;

begin
  temp:=Sprites;
  last:=nil;

  while (temp<>nil) and (temp<>Sprite) do
    begin
      last:=temp;
      temp:=temp.next;
    end;

  result:=last;
end;

{ Updates the positions and re-draws any dirty rectangles }
procedure TSpriteEngine.UpdateSprites;
var
  s,temp : TSprite;
  r      : TDirtyRect;
  rec    : TDirtyRect;
  x,y    : integer;

begin
  rec:=TDirtyRect.Create;
  r:=TDirtyRect.Create;

  try
    { Now look for any collisions between sprites }
    s:=fSprites.next;  // Skip the background
    while s<>nil do
      begin
        if s.visible then
          begin
            rec.assign(s.x,
                       s.y,
                       s.x+fScreen.Images[s.image].width,
                       s.y+fScreen.Images[s.image].height);

            temp:=fSprites.next;
            while temp<>nil do
              begin
                if (temp<>s) and (temp.mask and s.mask<>0) and temp.visible then
                  begin
                    r.assign(temp.x,
                             temp.y,
                             temp.x+fScreen.Images[temp.image].width,
                             temp.y+fScreen.Images[temp.image].height);

                    if r.intersects(rec) then
                      if fScreen.Images[s.image].
                                 CollisionTest(s.x,s.y,
                                               fScreen.Images[temp.image],
                                               temp.x,temp.y,TRUE) then
                      begin
                        s.collide(temp);

                        if assigned(fOnCollision) then
                          fOnCollision(self,s,temp);
                      end;
                  end;

                temp:=temp.next;
              end;
          end;
        s:=s.next;
      end;
  finally
    r.free;
  end;
  
  { Call the tick method of each sprite in the list }
  s:=fSprites;
  while s<>nil do
    begin
      s.tick;
      s:=s.next;
    end;

  { Now remove any sprites that have been killed from the list }
  s:=fSprites;
  while s<>nil do
    begin
      temp:=s.next;

      if (s.flags and hasBeenKilled)<>0 then
        s.free;

      s:=temp;
    end;

  { Now process the dirty rectangle list }
  try
    if dirty<>nil then
      begin
        MergeDirtyList;

        r:=dirty;
        while r<>nil do
          begin
            temp:=sprites;

            while temp<>nil do
              begin
                if temp.visible then
                  begin
                    if temp.image>=0 then
                      rec.assign(temp.x,
                                  temp.y,
                                  temp.x+fScreen.Images[temp.image].width,
                                  temp.y+fScreen.Images[temp.image].height)
                    else
                      rec.assign(0,0,fScreen.back.width,fScreen.back.height);

                    if r.intersects(rec) then
                      DrawSprite(temp,r);
                  end;

                temp:=temp.next;
              end;

            r:=r.next;
          end;

        EmptyDirtyList;
      end;
  finally
    rec.Free;
  end;
end;

constructor TSpriteEngine.Create(AOwner : TComponent);
begin
  inherited Create(AOwner);

  if not (csDesigning in ComponentState) then
    begin
      { Create the default black background }
      fBackground:=TBackground.Create(self,-1);
      fBackground.fVisible:=TRUE;
      AddDirtyRect(0,0,1024,1024);
    end;
end;

destructor TSpriteEngine.Destroy;

  procedure KillEm(s : TSprite); forward;
  procedure KillEm(s : TSprite);
  begin
    if s<>nil then
      begin
        if s.next<>nil then
          KillEm(s.next);

        s.free;
      end;
  end;

begin
  if not (csDesigning in ComponentState) then
    KillEm(fSprites);

  fSprites:=nil;

  inherited Destroy;
end;

constructor TSprite.Create(AOwner : TSpriteEngine);
var
  temp : TSprite;

begin
  inherited Create;

  Owner:=AOwner;
  x:=0;
  y:=0;
  Image:=0;
  visible:=FALSE;
  flags:=0;
  next:=nil;
  mask:=$ffffffff;  // Can collide with any sprite with a non-zero mask

  if AOwner.fScreen<>nil then
    bounds:=rect(0,0,AOwner.fScreen.ScreenWidth,
                    AOwner.fScreen.ScreenHeight);

  temp:=owner.BeforeSprite(nil);

  prev:=temp;

  if temp<>nil then
    temp.next:=self
  else
    owner.fSprites:=self;
end;

destructor TSprite.Destroy;
var
  temp : TSprite;

begin
  Hide;  // Make sure the screen is updated

  // Now remove the sprite from the list
  temp:=owner.BeforeSprite(self);

  if temp<>nil then
    begin
      temp.next:=next;
      if next<>nil then
        next.prev:=temp;
    end
  else
    owner.fSprites:=next;

  inherited Destroy;
end;

{ Move the sprite to a new location }
procedure TSprite.MoveTo(ax,ay : integer);
begin
  Invalidate;  // Add the current sprite area to the dirty rectangle list
  fx:=ax;
  fy:=ay;
  Invalidate;  // Add the new area to the dirty rectangle list
end;

{ Move the sprite relative to it's current position }
procedure TSprite.MoveRel(dx,dy : integer);
begin
  MoveTo(x+dx,y+dy);
end;

{ Add the area of the screen that is covered by the sprite to the dirty
  rectangle list }
procedure TSprite.Invalidate;
begin
  if visible then
    if image>=0 then
      owner.AddDirtyRect(fx,fy,Owner.fScreen.images[image].width,
                               Owner.fScreen.images[image].height)
    else
      owner.AddDirtyRect(0,0,owner.fScreen.ScreenWidth,
                             owner.fScreen.ScreenHeight);
end;

{ By default a sprite is not visible so a call to show is necessary to display
  the sprite on the screen }
procedure TSprite.Show;
begin
  if not fVisible then
    begin
      fVisible:=TRUE;
      Invalidate;
   end;
end;

{ Makes a sprite invisible.  Note, invisible sprites to not collide with other
  sprites }
procedure TSprite.Hide;
begin
  if fVisible then
    begin
      Invalidate;
      fVisible:=FALSE;
    end;
end;

{ Sets the sprite image.  This routine is access via the Image property }
procedure TSprite.SetImage(value : integer);
begin
  if visible and (value<>fImage) then
    begin
      Hide;
      fImage:=value;
      Show;
    end
  else
    fImage:=value;
end;

{ Sets various properties of the sprite }
procedure TSprite.SetInfo(index : integer; value : integer);
begin
  case index of
    1 : fx:=value;    { Possition }
    2 : fy:=value;
    3 : fvx:=value;   { Velovity }
    4 : fvy:=value;
    5 : fax:=value;   { Acceleration }
    6 : fay:=value;
  end;
end;

{ Returns various properties of the sprite }
function TSprite.GetInfo(index : integer):integer;
begin
  case index of
    1 : result:=fx;   { Position }
    2 : result:=fy;
    3 : result:=fvx;  { Velocity }
    4 : result:=fvy;
    5 : result:=fax;  { Acceleration }
    6 : result:=fay;
  end;
end;

{ Takes care of the visible property }
procedure TSprite.SetVisible(value : boolean);
begin
  if value then
    show
  else
    hide;
end;

{ This is called each frame by the UpdateSprites method of the Sprite Engine.
  Any code that is placed in the Tick method should be as quick as possible
  as it will most likely be called many times per frame.

  This method is used to add some intelligence to the sprite.  The behaviour of
  a sprite, by default is to move in a straight line }
procedure TSprite.Tick;
var
  e : TEdges;

begin
  vx:=vx+ax;
  vy:=vy+ay;

  if (vx<>0) and (vy<>0) then
    MoveRel(vx,vy);

  e:=[];

  if (x<=bounds.left) or
     (x+Owner.fScreen.Images[image].width>=bounds.right) then
    begin
      if vx<0 then
        e:=e+[LeftEdge]
      else
        e:=e+[RightEdge];

      if flags and wantBounce<>0 then
        begin
          vx:=-vx;
          MoveRel(vx,0);
        end;

      if flags and wantStick<>0 then
        begin
          vx:=0;

          if x<bounds.left then
            MoveTo(bounds.left,y);

          if x+Owner.fScreen.Images[image].width>bounds.right then
            MoveTo(bounds.right-Owner.fScreen.Images[image].width,y);
        end;
    end;

  if (y<=bounds.top) or
     (y+Owner.fScreen.Images[image].height>=bounds.bottom) then
    begin
      if vy<0 then
        e:=e+[TopEdge]
      else
        e:=e+[BottomEdge];

      if flags and wantBounce<>0 then
        begin
          vy:=-vy;
          MoveRel(0,vy);
        end;

      if flags and wantStick<>0 then
        begin
          vy:=0;

          if y<bounds.top then
            MoveTo(x,bounds.top);

          if y+Owner.fScreen.Images[image].height>bounds.bottom then
            MoveTo(x,bounds.bottom-Owner.fScreen.Images[image].height);
        end;
    end;

  if e<>[] then
    HitEdge(e);
end;

procedure TSprite.HitEdge(e : TEdges);
begin
end;

{ Called when a sprite collides with another sprite }
procedure TSprite.Collide(s : TSprite);
begin
end;

{ Set's the flag to kill the sprite }
procedure TSprite.Kill;
begin
  flags:=flags or HasBeenKilled;
end;

{ The TBackground class is just a special case of the TSprite class that takes
  image number in it's constructor.

  If this image number is negative it is used as an index into the palette to
  fill the background with.

  palette index = abs(image) - 1

  }
constructor TBackground.Create(AOwner : TSpriteEngine; aImage : integer);
begin
  inherited Create(AOwner);

  image:=aImage;
end;

procedure TFractSprite.Tick;
var
  e : TEdges;

begin
  rx:=rx+rvx;
  ry:=ry+rvy;

  MoveTo(round(rx),round(ry));
  e:=[];

  if (x<=bounds.left) or
     (x+Owner.fScreen.Images[image].width>=bounds.right) then
    begin
      if vx<0 then
        e:=e+[LeftEdge]
      else
        e:=e+[RightEdge];

      if flags and wantBounce<>0 then
        begin
          rvx:=-rvx;
          MoveRel(round(rvx),0);
        end;

      if flags and wantStick<>0 then
        begin
          rvx:=0;

          if x<bounds.left then
            MoveTo(bounds.left,y);

          if x+Owner.fScreen.Images[image].width>bounds.right then
            MoveTo(bounds.right-Owner.fScreen.Images[image].height,y);

          rx:=x;
        end;
    end;

  if (y<=bounds.top) or
     (y+Owner.fScreen.Images[image].height>=bounds.bottom) then
    begin
      if rvy<0 then
        e:=e+[TopEdge]
      else
        e:=e+[BottomEdge];

      if flags and wantBounce<>0 then
        begin
          rvy:=-rvy;
          MoveRel(0,round(rvy));
        end;

      if flags and wantStick<>0 then
        begin
          rvy:=0;

          if y<bounds.top then
            MoveTo(x,bounds.top);

          if y+Owner.fScreen.Images[image].height>bounds.bottom then
            MoveTo(x,bounds.bottom-Owner.fScreen.Images[image].height);

          ry:=y;
        end;
    end;

  if e<>[] then
    HitEdge(e);
end;

constructor TAnimSprite.Create(AOwner : TSpriteEngine; base,frames : integer);
begin
  inherited Create(AOwner);

  ImageBase:=base;
  numFrames:=frames;
  image:=base;
  speed:=1;
end;

procedure TAnimSprite.Tick;
begin
  inherited Tick;

  dec(count);
  if count<=0 then
    begin
      count:=Speed;

      if (image+1-imageBase)>=numFrames then
        image:=imageBase
      else
        image:=image+1;
    end;
end;

procedure Register;
begin
  registerComponents('DGC',[TSpriteEngine]);
end;

end.
